#-------------------------------------------------------------------------------
# Author: Lukasz Janyst <lukasz@jany.st>
# Date:   03.12.2017
#
# Licensed under the 3-Clause BSD License, see the LICENSE file for details.
#-------------------------------------------------------------------------------

"""
All the functionality running for running the jobs and keeping a proper record
of them.
"""

import tempfile
import pickle
import shutil
import os

from twisted.application.service import Service
from twisted.internet.defer import inlineCallbacks, returnValue
from twisted.internet.utils import getProcessValue, getProcessOutputAndValue
from twisted.internet.task import LoopingCall
from distutils.spawn import find_executable
from twisted.logger import Logger
from collections import namedtuple
from .schedule import Schedule, Job, Actor, Status
from schedule import Scheduler
from datetime import datetime
from .utils import schedule_job, run_process, twisted_sleep, exc_repr


#-------------------------------------------------------------------------------
Project = namedtuple('Project', ['name', 'archive', 'spiders'])
RunningJob = namedtuple('RunningJob', ['process', 'finished_d', 'time_started'])


#-------------------------------------------------------------------------------
class Controller(Service):
    """
    An object of this class is responsible for glueing together all the other
    components. It needs the following options in the `scrapy-do` section of
    the configuration:

      * `project-store` - a directory for all the data and metadata.
      * `job-slots` - number of jobs to run in parallel
      * `completed-cap` - number of completed jobs to keep while purging the old
        jobs

    :param config: A :class:`Config <scrapy_do.config.Config>`.
                   contains the following options in the `scrapy-do` section:
    """

    log = Logger()

    #---------------------------------------------------------------------------
    def __init__(self, config):
        #-----------------------------------------------------------------------
        # Configuration
        #-----------------------------------------------------------------------
        self.log.info('Creating controller')
        self.config = config
        ps = config.get_string('scrapy-do', 'project-store')
        ps_abs = os.path.join(os.getcwd(), ps)
        self.project_store = ps if ps.startswith('/') else ps_abs
        self.job_slots = config.get_int('scrapy-do', 'job-slots')
        self.completed_cap = config.get_int('scrapy-do', 'completed-cap')
        self.metadata_path = os.path.join(self.project_store, 'metadata.pkl')
        self.schedule_path = os.path.join(self.project_store, 'schedule.db')
        self.log_dir = os.path.join(self.project_store, 'log-dir')
        self.spider_data_dir = os.path.join(self.project_store, 'spider-data')
        self.running_jobs = {}
        self.scheduled_jobs = {}
        self.counter_run = 0
        self.counter_success = 0
        self.counter_failure = 0
        self.counter_cancel = 0
        self.start_time = datetime.now()

        #-----------------------------------------------------------------------
        # Create all the directories
        #-----------------------------------------------------------------------
        dirs = [self.project_store, self.log_dir, self.spider_data_dir]
        for d in dirs:
            try:
                os.makedirs(d)
            except FileExistsError:
                pass

        if os.path.exists(self.metadata_path):
            with open(self.metadata_path, 'rb') as f:
                self.projects = pickle.load(f)
        else:
            self.projects = {}
            with open(self.metadata_path, 'wb') as f:
                pickle.dump(self.projects, f)

        #-----------------------------------------------------------------------
        # Set the scheduler up
        #-----------------------------------------------------------------------
        self.schedule = Schedule(self.schedule_path)
        self.scheduler = Scheduler()

        for job in self.schedule.get_jobs(Status.SCHEDULED):
            self.log.info('Re-scheduling: {}'.format(str(job)))
            sch_job = schedule_job(self.scheduler, job.schedule)
            sch_job.do(lambda: self.schedule_job(job.project, job.spider, 'now',
                                                 Actor.SCHEDULER))
            self.scheduled_jobs[job.identifier] = sch_job

        #-----------------------------------------------------------------------
        # If we have any jobs marked as RUNNING in the schedule at this point,
        # it means that the daemon was killed while the jobs were running. We
        # mark these jobs as pending, so that they can be restarted as soon
        # as possible
        #-----------------------------------------------------------------------
        for job in self.schedule.get_jobs(Status.RUNNING):
            self.log.info('Restarting interrupted: {}'.format(str(job)))
            job.status = Status.PENDING
            self.schedule.commit_job(job)

        #-----------------------------------------------------------------------
        # Set up the service
        #-----------------------------------------------------------------------
        self.setName('Controller')
        self.scheduler_loop = LoopingCall(self.run_scheduler)
        self.crawlers_loop = LoopingCall(self.run_crawlers)
        self.purger_loop = LoopingCall(self.purge_completed_jobs)

    #---------------------------------------------------------------------------
    def startService(self):
        """
        Start the twisted related functionality.
        """
        self.log.info('Starting controller')
        self.scheduler_loop.start(1.)
        self.crawlers_loop.start(1.)
        self.purger_loop.start(10.)

    #---------------------------------------------------------------------------
    def stopService(self):
        """
        Stop the twisted related functionality.
        """
        self.log.info('Stopping controller')
        self.scheduler_loop.stop()
        self.crawlers_loop.stop()
        self.purger_loop.stop()
        return self.wait_for_running_jobs(cancel=True)

    #---------------------------------------------------------------------------
    @inlineCallbacks
    def push_project(self, name, data):
        """
        Register a project of a given name with the zipped code passed in data.

        :param name: Name of the project
        :param data: Binary blob with a zipped project code
        :return:     A deferred that gets called back with a list of spiders
                     provided by the project, of a `ValueError` failure.
        """
        self.log.info('Pushing project "{}"'.format(name))

        #-----------------------------------------------------------------------
        # Store the data in a temoporary file
        #-----------------------------------------------------------------------
        tmp = tempfile.mkstemp()
        with open(tmp[0], 'wb') as f:
            f.write(data)

        #-----------------------------------------------------------------------
        # Unzip to a temporary directory
        #-----------------------------------------------------------------------
        temp_dir = tempfile.mkdtemp()

        unzip = find_executable('unzip')
        ret_code = yield getProcessValue(unzip, args=(tmp[1],), path=temp_dir)
        if ret_code != 0:
            shutil.rmtree(temp_dir)
            os.remove(tmp[1])
            self.log.debug('Failed to unzip data using "{}"'.format(unzip))
            raise ValueError('Not a valid zip archive')

        #-----------------------------------------------------------------------
        # Figure out the list of spiders
        #-----------------------------------------------------------------------
        temp_proj_dir = os.path.join(temp_dir, name)
        if not os.path.exists(temp_proj_dir):
            shutil.rmtree(temp_dir)
            os.remove(tmp[1])
            raise ValueError('Project {} not found in the archive'.format(name))

        scrapy = find_executable('scrapy')
        ret = yield getProcessOutputAndValue(scrapy, ('list',),
                                             path=temp_proj_dir)
        out, err, ret_code = ret

        if ret_code != 0:
            shutil.rmtree(temp_dir)
            os.remove(tmp[1])
            raise ValueError('Unable to get the list of spiders')

        spiders = out.decode('utf-8').split()

        shutil.rmtree(temp_dir)

        #-----------------------------------------------------------------------
        # Move to the final position and store the matadata
        #-----------------------------------------------------------------------
        archive = os.path.join(self.project_store, name + '.zip')
        shutil.move(tmp[1], archive)
        prj = Project(name, archive, spiders)
        self.projects[name] = prj
        with open(self.metadata_path, 'wb') as f:
            pickle.dump(self.projects, f)

        self.log.info('Added project "{}" with spiders {}'.format(
            name, prj.spiders))
        returnValue(prj.spiders)

    #---------------------------------------------------------------------------
    def get_projects(self):
        """
        Get the names of all the registred projects.
        """
        return list(self.projects.keys())

    #---------------------------------------------------------------------------
    def get_spiders(self, project_name):
        """
        Get names of all the spiders in the project.

        :param project_name: Name of the project
        :raises KeyError:    If the project name is not known
        """
        if project_name not in self.projects.keys():
            raise KeyError('Unknown project ' + project_name)
        return self.projects[project_name].spiders

    #---------------------------------------------------------------------------
    def schedule_job(self, project, spider, when, actor=Actor.USER):
        """
        Schedule a crawler job.

        :param project: Name of the project
        :param spider:  Name of the spider
        :param when:    A scheduling spec as handled by :meth:`schedule_job
                        <scrapy_do.utils.schedule_job>`
        :param actor:   :data:`Actor <scrapy_do.schedule.Actor>` triggering the
                        event
        :return:        A string identifier of a job
        """
        if project not in self.projects.keys():
            raise KeyError('Unknown project ' + project)

        if spider not in self.projects[project].spiders:
            raise KeyError('Unknown spider {}/{}'.format(project, spider))

        job = Job(Status.PENDING, actor, 'now', project, spider)
        if when != 'now':
            sch_job = schedule_job(self.scheduler, when)
            sch_job.do(lambda: self.schedule_job(project, spider, 'now',
                                                 Actor.SCHEDULER))
            self.scheduled_jobs[job.identifier] = sch_job
            job.status = Status.SCHEDULED
            job.schedule = when

        self.log.info('Scheduling: {}'.format(str(job)))
        self.schedule.add_job(job)
        return job.identifier

    #---------------------------------------------------------------------------
    def get_jobs(self, job_status):
        """
        See :meth:`Schedule.get_jobs <scrapy_do.schedule.Schedule.get_jobs>`.
        """
        return self.schedule.get_jobs(job_status)

    #---------------------------------------------------------------------------
    def get_active_jobs(self):
        """
        See :meth:`Schedule.get_active_jobs
        <scrapy_do.schedule.Schedule.get_active_jobs>`.
        """
        return self.schedule.get_active_jobs()

    #---------------------------------------------------------------------------
    def get_completed_jobs(self):
        """
        See :meth:`Schedule.get_completed_jobs
        <scrapy_do.schedule.Schedule.get_completed_jobs>`.
        """
        return self.schedule.get_completed_jobs()

    #---------------------------------------------------------------------------
    def get_job(self, job_id):
        """
        See :meth:`Schedule.get_job <scrapy_do.schedule.Schedule.get_job>`.
        """
        return self.schedule.get_job(job_id)

    #---------------------------------------------------------------------------
    def run_scheduler(self):
        """
        Run the `schedule.Scheduler` jobs.
        """
        self.scheduler.run_pending()

    #---------------------------------------------------------------------------
    @inlineCallbacks
    def _run_crawler(self, project, spider, job_id):
        #-----------------------------------------------------------------------
        # Unzip to a temporary directory
        #-----------------------------------------------------------------------
        temp_dir = tempfile.mkdtemp()
        archive = os.path.join(self.project_store, project + '.zip')

        unzip = find_executable('unzip')
        ret_code = yield getProcessValue(unzip, args=(archive,), path=temp_dir)
        if ret_code != 0:
            shutil.rmtree(temp_dir)
            msg = 'Unable to unzip with {}. Archive corrupted?'.format(unzip)
            self.log.error(msg)
            raise IOError('Cannot unzip the project archive')

        #-----------------------------------------------------------------------
        # Run the crawler
        #-----------------------------------------------------------------------
        temp_proj_dir = os.path.join(temp_dir, project)
        env = {'SPIDER_DATA_DIR': self.spider_data_dir}
        process, finished = run_process('scrapy', ['crawl', spider], job_id,
                                        self.log_dir, env=env,
                                        path=temp_proj_dir)

        #-----------------------------------------------------------------------
        # Clean up
        #-----------------------------------------------------------------------
        def clean_up(status):
            shutil.rmtree(temp_dir)
            return status
        finished.addBoth(clean_up)

        returnValue((process, finished))

    #---------------------------------------------------------------------------
    def run_crawlers(self):
        """
        Spawn as many crawler processe out of pending jobs as there is free
        job slots.
        """
        jobs = self.schedule.get_jobs(Status.PENDING)
        jobs.reverse()
        while len(self.running_jobs) < self.job_slots and jobs:
            self.counter_run += 1
            #-------------------------------------------------------------------
            # Run the job
            #-------------------------------------------------------------------
            job = jobs.pop()
            job.status = Status.RUNNING
            self.schedule.commit_job(job)
            # Use a placeholder until the process is actually started, so that
            # we do not exceed the quota due to races.
            self.running_jobs[job.identifier] = None

            d = self._run_crawler(job.project, job.spider, job.identifier)

            #-------------------------------------------------------------------
            # Error starting the job
            #-------------------------------------------------------------------
            def spawn_errback(error, job):
                self.counter_failure += 1
                job.status = Status.FAILED
                self.schedule.commit_job(job)
                self.log.error('Unable to start job {}: {}'.format(
                    job.identifier, exc_repr(error.value)))
                del self.running_jobs[job.identifier]

            #-------------------------------------------------------------------
            # Job started successfully
            #-------------------------------------------------------------------
            def spawn_callback(value, job):
                # Put the process object and the finish deferred in the
                # dictionary
                running_job = RunningJob(value[0], value[1], datetime.now())
                self.running_jobs[job.identifier] = running_job
                self.log.info('Job {} started successfully'.format(
                    job.identifier))

                #---------------------------------------------------------------
                # Finish things up
                #---------------------------------------------------------------
                def finished_callback(exit_code):
                    if exit_code == 0:
                        self.counter_success += 1
                        job.status = Status.SUCCESSFUL
                    else:
                        self.counter_failure += 1
                        job.status = Status.FAILED

                    rj = self.running_jobs[job.identifier]
                    job.duration = (datetime.now() - rj.time_started).seconds
                    msg = "Job {} exited with code {}".format(job.identifier,
                                                              exit_code)
                    self.log.info(msg)
                    self.schedule.commit_job(job)
                    del self.running_jobs[job.identifier]
                    return exit_code

                value[1].addCallback(finished_callback)

            d.addCallbacks(spawn_callback, spawn_errback,
                           callbackArgs=(job,), errbackArgs=(job,))

    #---------------------------------------------------------------------------
    @inlineCallbacks
    def wait_for_starting_jobs(self):
        """
        Wait until all the crawling processes in the job slots started.

        :return: A deferred triggered when all the processes have started
        """
        num_starting = 1  # whatever to loop at least once
        while num_starting:
            num_starting = 0
            for k, v in self.running_jobs.items():
                if v is None:
                    num_starting += 1
            yield twisted_sleep(0.1)

    #---------------------------------------------------------------------------
    @inlineCallbacks
    def wait_for_running_jobs(self, cancel=False):
        """
        Wait for all the running jobs to finish.

        :param cancel: If `True` send a `SIGTERM` signal to each of the running
                       crawlers
        :return:       A deferred triggered when all the running jobs have
                       finished
        """
        yield self.wait_for_starting_jobs()

        #-----------------------------------------------------------------------
        # Send SIGTERM if requested
        #-----------------------------------------------------------------------
        if cancel:
            for job_id in self.running_jobs:
                rj = self.running_jobs[job_id]
                rj.process.signalProcess('TERM')

        #-----------------------------------------------------------------------
        # Wait for the jobs to finish
        #-----------------------------------------------------------------------
        to_finish = []
        for job_id in self.running_jobs:
            rj = self.running_jobs[job_id]
            to_finish.append(rj.finished_d)

        for d in to_finish:
            yield d

    #---------------------------------------------------------------------------
    @inlineCallbacks
    def cancel_job(self, job_id):
        """
        Cancel a job.

        :param job_id: A string identifier of a job
        :return:       A deferred that is triggered when the job is cancelled
        """
        job = self.schedule.get_job(job_id)
        self.log.info('Canceling: {}'.format(str(job)))

        #-----------------------------------------------------------------------
        # Scheduled
        #-----------------------------------------------------------------------
        if job.status == Status.SCHEDULED:
            job.status = Status.CANCELED
            self.schedule.commit_job(job)
            self.scheduler.cancel_job(self.scheduled_jobs[job_id])
            del self.scheduled_jobs[job_id]

        #-----------------------------------------------------------------------
        # Pending
        #-----------------------------------------------------------------------
        elif job.status == Status.PENDING:
            job.status = Status.CANCELED
            self.schedule.commit_job(job)

        #-----------------------------------------------------------------------
        # Running
        #-----------------------------------------------------------------------
        elif job.status == Status.RUNNING:
            while True:
                if job_id not in self.running_jobs:
                    raise KeyError('Job {} is not active'.format(job_id))
                if self.running_jobs[job_id] is None:
                    yield twisted_sleep(0.1)  # wait until the job starts
                else:
                    break
            rj = self.running_jobs[job_id]
            rj.process.signalProcess('TERM')
            yield rj.finished_d
            self.counter_failure -= 1
            self.counter_cancel += 1
            job.status = Status.CANCELED
            job.duration = (datetime.now() - rj.time_started).seconds
            self.schedule.commit_job(job)

        #-----------------------------------------------------------------------
        # Not active
        #-----------------------------------------------------------------------
        else:
            raise KeyError('Job {} is not active'.format(job_id))

    #---------------------------------------------------------------------------
    def purge_completed_jobs(self):
        """
        Purge all the old jobs exceeding the completed cap.
        """
        old_jobs = self.get_completed_jobs()[self.completed_cap:]

        if len(old_jobs):
            self.log.info('Purging {} old jobs'.format(len(old_jobs)))

        for job in old_jobs:
            self.schedule.remove_job(job.identifier)
            for log_type in ['.out', '.err']:
                log_file = os.path.join(self.log_dir, job.identifier + log_type)
                if os.path.exists(log_file):
                    os.remove(log_file)



#-------------------------------------------------------------------------------
# Author: Lukasz Janyst <lukasz@jany.st>
# Date:   26.11.2017
#
# Licensed under the 3-Clause BSD License, see the LICENSE file for details.
#-------------------------------------------------------------------------------

"""
A collection of utility classes and functions used throughout the project.
"""

import importlib
import os

from twisted.internet.protocol import ProcessProtocol
from twisted.internet.defer import Deferred
from twisted.internet import reactor, task
from distutils.spawn import find_executable
from datetime import datetime
from schedule import Job as SchJob


#-------------------------------------------------------------------------------
def exc_repr(e):
    """
    Return a string representation of an exception together with the excepion
    name.
    """
    return "{}: {}".format(type(e).__name__, str(e))


#-------------------------------------------------------------------------------
def get_object(name):
    """
    Retrieve an object from a module given its fully qualified name. For
    example: `get_object('scrapy_do.webservice.Status')`.
    """
    name = name.split('.')
    object_name = name[-1]
    module = importlib.import_module('.'.join(name[:-1]))
    return getattr(module, object_name)


#-------------------------------------------------------------------------------
class TimeStamper:
    """
    Set the timestamp attribute of the object whenever the associated attribute
    is set. For example:

    :Example:

        >>> class Test:
        >>>     attr = TimeStamper('_attr')
        >>>
        >>>     def __init__(self, attr):
        >>>         self._attr = attr
        >>>         self.timestamp = datetime.now()
        >>> test = Test('foo')
        >>> test.attr
        'foo'
        >>> test.timestamp
        datetime.datetime(2017, 12, 2, 23, 0, 56, 671634)
        >>> test.attr = 'bar'
        >>> test.timestamp
        datetime.datetime(2017, 12, 2, 23, 1, 9, 688899)
    """

    #---------------------------------------------------------------------------
    def __init__(self, attr_name):
        self.attr_name = attr_name

    #---------------------------------------------------------------------------
    def __get__(self, obj, obj_type):
        return getattr(obj, self.attr_name)

    #---------------------------------------------------------------------------
    def __set__(self, obj, value):
        obj.timestamp = datetime.now()
        return setattr(obj, self.attr_name, value)


#-------------------------------------------------------------------------------
def _build_directive_map(job):
    #---------------------------------------------------------------------------
    # A list of valid directives
    #---------------------------------------------------------------------------
    directive_names = ['second', 'seconds', 'minute', 'minutes', 'hour',
                       'hours', 'day', 'days', 'week', 'weeks', 'monday',
                       'tuesday', 'wednesday', 'thursday', 'friday',
                       'saturday', 'sunday', 'at', 'to']

    #---------------------------------------------------------------------------
    # Get an appropriate setter reference
    #---------------------------------------------------------------------------
    def get_attr(obj, attr):
        for obj in [obj] + obj.__class__.mro():
            if attr in obj.__dict__:
                ret = obj.__dict__[attr]
                if isinstance(ret, property):
                    return lambda x: ret.__get__(x, type(x))
                return ret

    #---------------------------------------------------------------------------
    # Build the dictionary of setters
    #---------------------------------------------------------------------------
    directive_map = {}
    for d in directive_names:
        directive_map[d] = get_attr(job, d)

    return directive_map


#-------------------------------------------------------------------------------
def _parse_args(directive, directives):
    #---------------------------------------------------------------------------
    # Check the argument to "to"
    #---------------------------------------------------------------------------
    if directive == 'to':
        arg = directives.pop()
        try:
            arg = int(arg)
        except ValueError:
            raise ValueError('The "to" directive expects an integer')
        return [arg]

    #---------------------------------------------------------------------------
    # Check the argument to "at"
    #---------------------------------------------------------------------------
    if directive == 'at':
        arg = directives.pop()
        arg_split = arg.split(':')

        if len(arg_split) != 2:
            raise ValueError('The "at" directive expects a string like "12:34"')

        try:
            int(arg_split[0])
            int(arg_split[1])
        except ValueError:
            raise ValueError('The "at" directive expects a string like "12:34"')

        return [arg]

    #---------------------------------------------------------------------------
    # Nothing else accepts arguments
    #---------------------------------------------------------------------------
    return []


#-------------------------------------------------------------------------------
def _parse_spec(job, spec):
    #---------------------------------------------------------------------------
    # Check the directive
    #---------------------------------------------------------------------------
    directives = spec.lower().split()

    if len(directives) < 2:
        raise ValueError('Spec too short')

    if directives[0] != 'every':
        raise ValueError('Spec must start with "every"')

    #---------------------------------------------------------------------------
    # Set up the interval if necessary
    #---------------------------------------------------------------------------
    try:
        interval = int(directives[1])
        job.interval = interval

        if len(directives) < 3:
            raise ValueError("Spec to short")
        directives = directives[2:]
    except ValueError:
        directives = directives[1:]

    #---------------------------------------------------------------------------
    # Parse the spec
    #---------------------------------------------------------------------------
    directive_map = _build_directive_map(job)
    directives.reverse()
    while directives:
        directive = directives.pop()
        if directive not in directive_map:
            raise ValueError('Unknown directive: ' + directive)

        args = _parse_args(directive, directives)

        try:
            directive_map[directive](job, *args)
        except AssertionError as e:
            raise ValueError(str(e))

    return job


#-------------------------------------------------------------------------------
def schedule_job(scheduler, spec):
    """
    Take a `schedule.Scheduler` object and an interval spec and convert it
    to a `schedule.Job` registered with the scheduler. The spec can be
    any string that can be translated to `schedule calls
    <https://schedule.readthedocs.io/en/stable/>`_. For example: string
    'every 2 to 3 minutes' corresponds to `schedule.every(2).to(3).minutes`.

    :param scheduler:   A `schedule.Scheduler`
    :param spec:        String containing the interval spec
    :return:            A `schedule.Job` registered with the scheduler
    :raises ValueError: If the spec is not a valid sequence of `schedule`
                        method calls
    """
    job = SchJob(1, scheduler)
    try:
        _parse_spec(job, spec)
    except Exception:
        scheduler.cancel_job(job)
        raise
    return job


#-------------------------------------------------------------------------------
def arg_require_all(dict, args):
    """
    Check if all of the args are present in the dict.

    :raises KeyError: If any argument is missing from the dict.
    """
    for arg in args:
        if arg not in dict:
            raise KeyError('Missing argument "{}"'.format(arg))


#-------------------------------------------------------------------------------
def arg_require_any(dict, args):
    """
    Check if any of the args is in the dict.

    :raises KeyError: If none of the args is present in the dict.
    """
    for arg in args:
        if arg in dict:
            return
    raise KeyError('Neither argument present: "{}"'.format(str(args)))


#-------------------------------------------------------------------------------
def twisted_sleep(time):
    """
    Return a deferred that will be triggered after the specified amount of
    time passes
    """
    return task.deferLater(reactor, time, lambda: None)


#-------------------------------------------------------------------------------
class LoggedProcessProtocol(ProcessProtocol):
    """
    An implementation of ProcessProtocol that forwards the program output
    to logfiles. It creates files `job_name.out` and `job_name.err` and
    redirects the standard output and standard error output of the
    process to the respective file. If a log file is empty upon program
    exit it is deleted. The :data:`finished <LoggedProcessProtocol.finished>`
    deferred is triggered upon process exit and called with it's exit code.

    :param job_name: Name of the job
    :param log_dir:  A directory to put the log files in
    """

    #---------------------------------------------------------------------------
    def __init__(self, job_name, log_dir):
        self.finished = Deferred()
        self.out_path = os.path.join(log_dir, job_name + '.out')
        self.err_path = os.path.join(log_dir, job_name + '.err')
        flags = os.O_WRONLY | os.O_CREAT | os.O_TRUNC
        self.out_fd = os.open(self.out_path, flags, 0o644)
        self.err_fd = os.open(self.err_path, flags, 0o644)
        os.set_inheritable(self.out_fd, True)
        os.set_inheritable(self.err_fd, True)

    #---------------------------------------------------------------------------
    def processExited(self, status):
        """
        Callback called by `twisted` upon process exit.
        """
        out_size = os.fstat(self.out_fd).st_size
        err_size = os.fstat(self.err_fd).st_size
        os.close(self.out_fd)
        os.close(self.err_fd)
        if out_size == 0:
            os.remove(self.out_path)
        if err_size == 0:
            os.remove(self.err_path)

        self.finished.callback(status.value.exitCode)


#-------------------------------------------------------------------------------
def run_process(cmd, args, job_name, log_dir, env=None, path=None):
    """
    Run a process using :class:`LoggedProcessProtocol <LoggedProcessProtocol>`

    :param cmd:      Command to run
    :param args:     Argument passed to the command
    :param job_name: Name of the job that will be used for the name of the
                     log files
    :param log_dir:  Directory where the log files will be stored
    :param env:      A dictionary with environment variables and their values
    :param path:     Program's working directory

    :return:         A tuple of an `IProcessTransport` object as returned
                     by twisted's `reactor.spawnProcess` and a deferred
                     called on program exit with the return code of the
                     process.
    """
    cmd = find_executable(cmd)
    args = [cmd] + args
    pp = LoggedProcessProtocol(job_name, log_dir)
    p = reactor.spawnProcess(pp, cmd, args, env=env, path=path,
                             childFDs={1: pp.out_fd, 2: pp.err_fd})
    return p, pp.finished


#-------------------------------------------------------------------------------
def pprint_relativedelta(delta):
    """
    Return a string representation of a relativedelta object in the form
    similar to: "1y 2m 3d 5h 6m". If any of the components is equal to zero,
    it's omitted.
    """
    ret = ''
    if delta.years:
        ret += '{}y '.format(delta.years)
    if delta.months:
        ret += '{}m '.format(delta.months)
    if delta.days:
        ret += '{}d '.format(delta.days)
    if delta.hours:
        ret += '{}h '.format(delta.hours)
    if delta.minutes:
        ret += '{}m '.format(delta.minutes)
    ret += '{}s'.format(delta.seconds)
    return ret



#-------------------------------------------------------------------------------
# Author: Lukasz Janyst <lukasz@jany.st>
# Date:   25.11.2017
#
# Licensed under the 3-Clause BSD License, see the LICENSE file for details.
#-------------------------------------------------------------------------------

from configparser import ConfigParser, NoSectionError, NoOptionError
from pkgutil import get_data


#-------------------------------------------------------------------------------
class Config:
    """
    A configuration dictionary with defaults. It's a wrapper around the
    :py:class:`configparser.ConfigParser`.
    """

    #---------------------------------------------------------------------------
    def __init__(self, config_files=[], package=__package__):
        """
        Load the default configuration from the specified package

        :param config_file: A list of paths to configuration files
        :param package: Name of the package in which to search for a default
                        config file; the current package is assumed if `None`
        """
        self.conf = ConfigParser()
        default_config = get_data(package, 'default.conf').decode('utf-8')
        self.conf.read_string(default_config)
        for config_file in config_files:
            self.conf.read(config_file)

    #---------------------------------------------------------------------------
    def __get_with_type(self, getter, section, option, default):
        """
        Get a an option of a specific type.
        """
        try:
            return getter(section, option)
        except (NoSectionError, NoOptionError, ValueError):
            if default is not None:
                return default
            raise

    #---------------------------------------------------------------------------
    def get_bool(self, section, option, default=None):
        """
        Get option value and coerce to boolean.

        :raises ValueError: value can not be coerced and no default is given
        :raises NoOptionError: option is missing and no default is given
        :raises NoSectionError: section is missing and no default is given
        """
        return self.__get_with_type(self.conf.getboolean, section, option,
                                    default)

    #---------------------------------------------------------------------------
    def get_int(self, section, option, default=None):
        """
        Get option value and coerce to integer. For details see
        :meth:`get_bool <Config.get_bool>`.
        """
        return self.__get_with_type(self.conf.getint, section, option, default)

    #---------------------------------------------------------------------------
    def get_float(self, section, option, default=None):
        """
        Get option value and coerce to float. For details see
        :meth:`get_bool <Config.get_bool>`.
        """
        return self.__get_with_type(self.conf.getfloat, section, option,
                                    default)

    #---------------------------------------------------------------------------
    def get_string(self, section, option, default=None):
        """
        Get option value as string.

        :raises NoOptionError: option is missing and no default is given
        :raises NoSectionError: section is missing and no default is given
        """
        return self.__get_with_type(self.conf.get, section, option, default)

    #---------------------------------------------------------------------------
    def get_options(self, section):
        """
        Get all options in a given section
        """
        return self.conf.items(section)



#-------------------------------------------------------------------------------
# Author: Lukasz Janyst <lukasz@jany.st>
# Date:   02.12.2017
#
# Licensed under the 3-Clause BSD License, see the LICENSE file for details.
#-------------------------------------------------------------------------------

"""
Functionality related to the database of jobs.
"""

import dateutil.parser
import sqlite3
import uuid

from scrapy_do.utils import TimeStamper
from datetime import datetime
from enum import Enum


#-------------------------------------------------------------------------------
class Status(Enum):
    """
    Status of the job.
    """
    SCHEDULED = 1
    PENDING = 2
    RUNNING = 3
    CANCELED = 4
    SUCCESSFUL = 5
    FAILED = 6


#-------------------------------------------------------------------------------
class Actor(Enum):
    """
    An entity responsible for job creation.
    """
    SCHEDULER = 1
    USER = 2


#-------------------------------------------------------------------------------
class Job:
    """
    A bin for all the parameters of a job.
    """

    status = TimeStamper('_status')
    actor = TimeStamper('_actor')
    schedule = TimeStamper('_schedule')
    project = TimeStamper('_project')
    spider = TimeStamper('_spider')
    duration = TimeStamper('_duration')

    #---------------------------------------------------------------------------
    def __init__(self, status=None, actor=None, schedule=None,
                 project=None, spider=None, timestamp=None, duration=None):
        self.identifier = str(uuid.uuid4())
        self._status = status
        self._actor = actor
        self._schedule = schedule
        self._project = project
        self._spider = spider
        self.timestamp = timestamp or datetime.now()
        self._duration = duration

    #---------------------------------------------------------------------------
    def __str__(self):
        s = 'Job[id="{}", actor="{}", schedule="{}", project="{}", spider="{}"]'
        s = s.format(self.identifier, self.actor.name, self.schedule,
                     self.project, self.spider)
        return s

    #---------------------------------------------------------------------------
    def to_dict(self):
        """
        Return all the parameters of the job as a dictionary
        """
        d = {
            'identifier': self.identifier,
            'status': self.status.name,
            'actor': self.actor.name,
            'schedule': self.schedule,
            'project': self.project,
            'spider': self.spider,
            'timestamp': str(self.timestamp),
            'duration': self.duration
        }
        return d


#-------------------------------------------------------------------------------
def _record_to_job(x):
    job = Job(status=Status(x[1]), actor=Actor(x[2]), schedule=x[3],
              project=x[4], spider=x[5], timestamp=dateutil.parser.parse(x[6]),
              duration=x[7])
    job.identifier = x[0]
    return job


#-------------------------------------------------------------------------------
class Schedule:
    """
    A persistent database of jobs.

    :param database: A file name where the database will be stored
    :param table:    Name of the table containing the schedule
    """

    #---------------------------------------------------------------------------
    def __init__(self, database=None, table='schedule'):
        #-----------------------------------------------------------------------
        # Create the database and the main table
        #-----------------------------------------------------------------------
        self.database = database or ':memory:'
        self.table = table
        self.db = sqlite3.connect(self.database,
                                  detect_types=sqlite3.PARSE_DECLTYPES)

        query = "CREATE TABLE IF NOT EXISTS {table} (" \
                "identifier VARCHAR(36) PRIMARY KEY, " \
                "status INTEGER NOT NULL, " \
                "actor INTEGER NOT NULL, " \
                "schedule VARCHAR(255), " \
                "project VARCHAR(255) NOT NULL, " \
                "spider VARCHAR(255) NOT NULL, " \
                "timestamp DATETIME NOT NULL, " \
                "duration INTEGER" \
                ")"
        query = query.format(table=self.table)
        self.db.execute(query)

        #-----------------------------------------------------------------------
        # Create the metadata table
        #-----------------------------------------------------------------------
        query = 'CREATE TABLE IF NOT EXISTS schedule_metadata (' \
                'key VARCHAR(255) PRIMARY KEY ON CONFLICT IGNORE, ' \
                'value VARCHAR(255) NOT NULL ' \
                ')'
        self.db.execute(query)

        query = 'INSERT INTO schedule_metadata ' \
                '(key, value) values ("version", "1") '
        self.db.execute(query)
        self.db.commit()

    #---------------------------------------------------------------------------
    def get_metadata(self, key):
        """
        Retrieve the matadata info with a given key
        """
        query = "SELECT * FROM schedule_metadata WHERE key=?"
        response = self.db.execute(query, (key, ))
        response = dict(response)
        return response[key]

    #---------------------------------------------------------------------------
    def get_jobs(self, job_status):
        """
        Retrieve a list of jobs with a given status

        :param job_status: One of :class:`statuses <Status>`
        """
        query = "SELECT * FROM {table} WHERE status=? ORDER BY timestamp DESC"
        query = query.format(table=self.table)
        response = self.db.execute(query, (job_status.value, ))
        return [_record_to_job(rec) for rec in response]

    #---------------------------------------------------------------------------
    def get_active_jobs(self):
        """
        Retrieve all the active jobs. Ie. all the jobs whose status is one of
        the following: :data:`SCHEDULED <Status.SCHEDULED>`,
        :data:`PENDING <Status.PENDING>`, or :data:`RUNNING <Status.RUNNING>`.
        """
        query = "SELECT * FROM {table} WHERE " \
                "status=1 OR status=2 OR status=3 "\
                "ORDER BY timestamp DESC"
        query = query.format(table=self.table)
        response = self.db.execute(query)
        return [_record_to_job(rec) for rec in response]

    #---------------------------------------------------------------------------
    def get_completed_jobs(self):
        """
        Retrieve all the completed jobs. Ie. all the jobs whose status is one of
        the  following: :data:`SUCCESSFUL <Status.SUCCESSFUL>`,
        :data:`FAILED <Status.FAILED>`, or :data:`CANCELED <Status.CANCELED>`.
        """
        query = "SELECT * FROM {table} WHERE " \
                "status=4 OR status=5 OR status=5 "\
                "ORDER BY timestamp DESC"
        query = query.format(table=self.table)
        response = self.db.execute(query)
        return [_record_to_job(rec) for rec in response]

    #---------------------------------------------------------------------------
    def get_job(self, identifier):
        """
        Retrieve a job by id

        :param identifier: A string identifier of the job
        """
        query = "SELECT * FROM {table} WHERE identifier=?"
        query = query.format(table=self.table)
        response = self.db.execute(query, (identifier, ))
        rec = response.fetchone()
        if rec is None:
            raise KeyError('No such job: "{}"'.format(identifier))
        return _record_to_job(rec)

    #---------------------------------------------------------------------------
    def add_job(self, job):
        """
        Add a job to the database

        :param job: A :class:`Job <Job>` object
        """
        query = "INSERT INTO {table}" \
                "(identifier, status, actor, schedule, project, spider, " \
                "timestamp, duration) " \
                "values (?, ?, ?, ?, ?, ?, ?, ?)"
        query = query.format(table=self.table)
        self.db.execute(query, (job.identifier, job.status.value,
                                job.actor.value, job.schedule, job.project,
                                job.spider, job.timestamp, job.duration))
        self.db.commit()

    #---------------------------------------------------------------------------
    def commit_job(self, job):
        """
        Modify an existing job

        :param job: A :class:`Job <Job>` object
        """
        query = "REPLACE INTO {table}" \
                "(identifier, status, actor, schedule, project, spider, " \
                "timestamp, duration) " \
                "values (?, ?, ?, ?, ?, ?, ?, ?)"
        query = query.format(table=self.table)
        self.db.execute(query, (job.identifier, job.status.value,
                                job.actor.value, job.schedule, job.project,
                                job.spider, job.timestamp, job.duration))
        self.db.commit()

    #---------------------------------------------------------------------------
    def remove_job(self, job_id):
        """
        Remove a job from the database

        :param identifier: A string identifier of the job
        """
        query = "DELETE FROM {table} WHERE identifier=?"
        query = query.format(table=self.table)
        self.db.execute(query, (job_id,))
        self.db.commit()



#-------------------------------------------------------------------------------
# Author: Lukasz Janyst <lukasz@jany.st>
# Date:   26.11.2017
#
# Licensed under the 3-Clause BSD License, see the LICENSE file for details.
#-------------------------------------------------------------------------------

import logging
import os

from twisted.application.internet import TCPServer, SSLServer
from twisted.application.service import MultiService, IServiceMaker
from twisted.internet.ssl import DefaultOpenSSLContextFactory
from twisted.python import log, usage
from zope.interface import implementer
from twisted.web import server

from .webservice import get_web_app
from .controller import Controller
from .config import Config
from .utils import exc_repr


#-------------------------------------------------------------------------------
class ScrapyDoOptions(usage.Options):
    optParameters = [
        ['config', 'c', '~/scrapy-do/config', 'A configuration file to load'],
    ]


#-------------------------------------------------------------------------------
@implementer(IServiceMaker)
class ScrapyDoServiceMaker():

    tapname = "scrapy-do"
    description = "A service running scrapy spiders."
    options = ScrapyDoOptions

    #---------------------------------------------------------------------------
    def _validate_web_config(self, config):
        interface = config.get_string('web', 'interface')
        port = config.get_int('web', 'port')
        https = config.get_bool('web', 'https')
        auth = config.get_bool('web', 'auth')
        key_file = None
        cert_file = None
        auth_file = None
        files_to_check = []

        if https:
            key_file = config.get_string('web', 'key')
            cert_file = config.get_string('web', 'cert')
            files_to_check += [key_file, cert_file]

        if auth:
            auth_file = config.get_string('web', 'auth-db')
            files_to_check.append(auth_file)

        for path in files_to_check:
            if not os.path.exists(path):
                raise FileNotFoundError(
                    "No such file or directory: '{}'".format(path))

        return interface, port, https, key_file, cert_file, auth, auth_file

    #---------------------------------------------------------------------------
    def _configure_web_server(self, config, controller):
        interface, port, https, key_file, cert_file, _, _ = \
            self._validate_web_config(config)

        site = server.Site(get_web_app(config, controller))
        if https:
            context = DefaultOpenSSLContextFactory(key_file, cert_file)
            web_server = SSLServer(port, site, context, interface=interface)
            method = 'https'
        else:
            web_server = TCPServer(port, site, interface=interface)
            method = 'http'

        log.msg(format="Scrapy-Do web interface is available at "
                       "%(method)s://%(interface)s:%(port)s/",
                method=method, interface=interface, port=port)

        return web_server

    #---------------------------------------------------------------------------
    def makeService(self, options):
        top_service = MultiService()
        config_file = os.path.expanduser(options['config'])
        config = Config([config_file])

        #-----------------------------------------------------------------------
        # Set up the controller
        #-----------------------------------------------------------------------
        try:
            controller = Controller(config)
            controller.setServiceParent(top_service)
        except Exception as e:
            log.msg(format="Unable to set up the controller: %(reason)s",
                    reason=exc_repr(e), logLevel=logging.ERROR)
            return top_service

        #-----------------------------------------------------------------------
        # Set up the web server
        #-----------------------------------------------------------------------
        try:
            web_server = self._configure_web_server(config, controller)
            web_server.setServiceParent(top_service)
        except Exception as e:
            log.msg(format="Scrapy-Do web interface could not have been "
                           "configured: %(reason)s",
                    reason=exc_repr(e), logLevel=logging.ERROR)
            return top_service

        return top_service



#-------------------------------------------------------------------------------
# Author: Lukasz Janyst <lukasz@jany.st>
# Date:   26.11.2017
#
# Licensed under the 3-Clause BSD License, see the LICENSE file for details.
#-------------------------------------------------------------------------------

import os
import json
import time
import psutil
import socket

from dateutil.relativedelta import relativedelta
from twisted.internet.defer import inlineCallbacks
from twisted.cred.checkers import FilePasswordDB
from twisted.web.resource import IResource
from twisted.cred.portal import IRealm, Portal
from twisted.web.static import File
from twisted.web.server import NOT_DONE_YET
from twisted.web.guard import HTTPAuthSessionWrapper, DigestCredentialFactory
from scrapy_do.utils import get_object
from zope.interface import implementer
from twisted.web import resource
from .schedule import Status as JobStatus
from datetime import datetime
from .utils import arg_require_all, arg_require_any, pprint_relativedelta


#-------------------------------------------------------------------------------
class WebApp(resource.Resource):

    #---------------------------------------------------------------------------
    def __init__(self, config, controller):
        super(WebApp, self).__init__()
        self.config = config
        self.controller = controller
        self.putChild(b'', Home())

        web_modules = config.get_options('web-modules')

        for mod_name, mod_class_name in web_modules:
            mod_class = get_object(mod_class_name)
            self.putChild(mod_name.encode('utf-8'), mod_class(self))


#-------------------------------------------------------------------------------
class Home(resource.Resource):

    isLeaf = True

    #---------------------------------------------------------------------------
    def render_GET(self, request):
        return "<html>Hello, world!</html>".encode('utf-8')


#-------------------------------------------------------------------------------
class JsonResource(resource.Resource):

    isLeaf = True

    #---------------------------------------------------------------------------
    def __init__(self, parent):
        super(JsonResource, self).__init__()
        self.parent = parent

    #---------------------------------------------------------------------------
    def render_json(self, request, data):
        json_data = json.dumps(data) + '\n'
        json_data = json_data.encode('utf-8')
        request.setHeader('Content-Type', 'application/json')
        request.setHeader('Content-Length', len(json_data))
        return json_data

    #---------------------------------------------------------------------------
    def render(self, request):
        try:
            data = super(JsonResource, self).render(request)
            if data == NOT_DONE_YET:
                return data
            data = {
                **{'status': 'ok'},
                **data
            }
            return self.render_json(request, data)
        except Exception as e:
            request.setResponseCode(400)
            data = {
                'status': 'error',
                'msg': str(e)
            }
            return self.render_json(request, data)


#-------------------------------------------------------------------------------
class Status(JsonResource):

    #---------------------------------------------------------------------------
    def render_GET(self, request):
        p = psutil.Process(os.getpid())
        controller = self.parent.controller
        uptime = relativedelta(datetime.now(), controller.start_time)
        resp = {
            'memory-usage': float(p.memory_info().rss) / 1024. / 1024.,
            'cpu-usage': p.cpu_percent(),
            'time': str(datetime.now()),
            'timezone': "{}; {}".format(time.tzname[0], time.tzname[1]),
            'hostname': socket.gethostname(),
            'uptime': pprint_relativedelta(uptime),
            'jobs-run': controller.counter_run,
            'jobs-successful': controller.counter_success,
            'jobs-failed': controller.counter_failure,
            'jobs-canceled': controller.counter_cancel
        }
        return resp


#-------------------------------------------------------------------------------
class PushProject(JsonResource):

    #---------------------------------------------------------------------------
    def render_POST(self, request):
        @inlineCallbacks
        def do_async():
            try:
                name = request.args[b'name'][0].decode('utf-8')
                data = request.args[b'archive'][0]
            except KeyError as e:
                result = {
                    'status': 'error',
                    'msg': 'Missing argument: ' + str(e)
                }
                request.setResponseCode(400)
                request.write(self.render_json(request, result))
                request.finish()
                return

            try:
                controller = self.parent.controller

                spiders = yield controller.push_project(name, data)
                result = {'status': 'ok', 'spiders': spiders}
            except Exception as e:
                request.setResponseCode(400)
                result = {'status': 'error', 'msg': str(e)}

            request.write(self.render_json(request, result))
            request.finish()
        do_async()
        return NOT_DONE_YET


#-------------------------------------------------------------------------------
class ListProjects(JsonResource):

    #---------------------------------------------------------------------------
    def render_GET(self, request):
        projects = self.parent.controller.get_projects()
        return {'projects': projects}


#-------------------------------------------------------------------------------
class ListSpiders(JsonResource):

    #---------------------------------------------------------------------------
    def render_GET(self, request):
        arg_require_all(request.args, [b'project'])
        project = request.args[b'project'][0].decode('utf-8')

        spiders = self.parent.controller.get_spiders(project)
        return {'project': project, 'spiders': spiders}


#-------------------------------------------------------------------------------
class ScheduleJob(JsonResource):

    #---------------------------------------------------------------------------
    def render_POST(self, request):
        arg_require_all(request.args, [b'project', b'spider', b'when'])
        project = request.args[b'project'][0].decode('utf-8')
        spider = request.args[b'spider'][0].decode('utf-8')
        when = request.args[b'when'][0].decode('utf-8')

        job_id = self.parent.controller.schedule_job(project, spider, when)
        return {'identifier': job_id}


#-------------------------------------------------------------------------------
class ListJobs(JsonResource):

    #---------------------------------------------------------------------------
    def render_GET(self, request):
        arg_require_any(request.args, [b'status', b'id'])
        if b'status' in request.args:
            status = request.args[b'status'][0].decode('utf-8')
            if status == 'ACTIVE':
                jobs = self.parent.controller.get_active_jobs()
            elif status == 'COMPLETED':
                jobs = self.parent.controller.get_completed_jobs()
            else:
                status = JobStatus[status]
                jobs = self.parent.controller.get_jobs(status)
        else:
            identifier = request.args[b'id'][0].decode('utf-8')
            jobs = [self.parent.controller.get_job(identifier)]

        return {'jobs': [job.to_dict() for job in jobs]}


#-------------------------------------------------------------------------------
class CancelJob(JsonResource):

    #---------------------------------------------------------------------------
    def render_POST(self, request):
        @inlineCallbacks
        def do_async():
            try:
                job_id = request.args[b'id'][0].decode('utf-8')
            except KeyError as e:
                result = {
                    'status': 'error',
                    'msg': 'Missing argument: ' + str(e)
                }
                request.setResponseCode(400)
                request.write(self.render_json(request, result))
                request.finish()
                return

            try:
                controller = self.parent.controller

                yield controller.cancel_job(job_id)
                result = {'status': 'ok'}
            except Exception as e:
                request.setResponseCode(400)
                result = {'status': 'error', 'msg': str(e)}

            request.write(self.render_json(request, result))
            request.finish()
        do_async()
        return NOT_DONE_YET


#-------------------------------------------------------------------------------
class GetLog(resource.Resource):

    isLeaf = False

    #---------------------------------------------------------------------------
    def __init__(self, parent):
        super(GetLog, self).__init__()
        data = File(parent.controller.log_dir, defaultType='text/plain')
        self.putChild(b'data', data)


#-------------------------------------------------------------------------------
@implementer(IRealm)
class PublicHTMLRealm:

    #---------------------------------------------------------------------------
    def __init__(self, config, controller):
        super(PublicHTMLRealm, self).__init__()
        self.config = config
        self.controller = controller

    #---------------------------------------------------------------------------
    def requestAvatar(self, avatar_id, mind, *interfaces):
        if IResource in interfaces:
            return (IResource, WebApp(self.config, self.controller),
                    lambda: None)
        raise NotImplementedError()


#-------------------------------------------------------------------------------
def get_web_app(config, controller):
    auth = config.get_bool('web', 'auth', False)
    if auth:
        auth_file = config.get_string('web', 'auth-db')
        portal = Portal(PublicHTMLRealm(config, controller),
                        [FilePasswordDB(auth_file)])
        credential_factory = DigestCredentialFactory('md5', b'scrapy-do')
        resource = HTTPAuthSessionWrapper(portal, [credential_factory])
        return resource

    return WebApp(config, controller)



from setuptools import setup

with open('README.rst') as readme:
    long_description = readme.read()

setup(
    name = 'scrapy-do',
    version = '0.1.0',
    author = 'Lukasz Janyst',
    author_email = 'xyz@jany.st',
    url = 'https://jany.st/scrapy-do.html',
    description = 'Spider Runner for Scrapy',
    long_description = long_description,
    license = 'BSD License',
    packages = ['scrapy_do'],
    include_package_data = True,
    package_data={
        '': ['*.conf'],
    },
    scripts=['scrapy-do'],
    classifiers = [
        'Framework :: Scrapy',
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Topic :: Internet :: WWW/HTTP',
        'Environment :: Console',
        'Environment :: No Input/Output (Daemon)',
    ],
    install_requires = [
        'scrapy', 'twisted', 'pyOpenSSL', 'psutil', 'python-dateutil',
        'schedule'
    ]
)
